@@ -1,9 +1,9 @@
-package RevML::Doctype::v0_27 ;
+package RevML::Doctype::v0_26 ;
 
 ##
 ## THIS FILE CREATED AUTOMATICALLY: YOU MAY LOSE ANY EDITS IF YOU MOFIFY IT.
 ##
-## When: Thu Sep  7 11:35:49 2000
+## When: Fri Aug 18 22:39:26 2000
 ## By:   RevML::Doctype, v0.1, (XML::Doctype, v0.11)
 ##
 
@@ -34,7 +34,7 @@ $doctype = bless( [
         'PATHS' => 6
       },
       undef,
-      '^<name><type>(?:<rev_id>(?:<change_id>)?<digest>|(?:<cvs_info>|<p4_info>|<source_safe_info>|<pvcs_info>)?(?:<branch_id>)?<rev_id>(?:<change_id>)?<time>(?:<mod_time>)?<user_id>(?:<p4_action>|<sourcesafe_action>)?(?:<label>)*(?:<lock>)?(?:<comment>)?(?:<delete>|<move>|(?:<content>|(?:<base_name>)?<base_rev_id><delta>)<digest>))$',
+      '^<name><type>(?:<cvs_info>|<p4_info>|<source_safe_info>|<pvcs_info>)?(?:<branch_id>)?<rev_id>(?:<change_id>)?<time>(?:<mod_time>)?<user_id>(?:<p4_action>|<sourcesafe_action>)?(?:<label>)*(?:<lock>)?(?:<comment>)?(?:<delete>|<move>|(?:<content>|(?:<base_name>)?<base_rev_id><delta>)<digest>)$',
       1,
       'rev',
       [
@@ -261,7 +261,7 @@ $doctype = bless( [
       {
         'version' => bless( [
           {},
-          '0.27',
+          '0.26',
           'version',
           undef,
           '#FIXED',
